#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include "TrailStack.h"
#include "Search.h"

using namespace std;

#define SEARCH_DBG 0

extern char prot_name[10];
extern int n; // number of variables
extern char primary[1000]; // sequence
extern double** dist;
extern int refFCC[1000][3];

extern long leaves;       // leaves visited
extern long expanded_nodes;       // nodes expanded
extern long pruned_nodes;       // nodes pruned

extern int* var_on_lev;  /// for each node, stores the variable currently labelled
extern AllVariables* nodes; /// nodes of the tree (array of variables)
extern AllVariables* temp_nodes; /// temp store for search_best_guided

extern TStack trailstack;

extern double best_energy;
extern AllVariables* best_sol;


extern long fail_cons;
extern long fail_prop;


extern int directions[12][3]; // codes every lattice direction 
extern int direct_indomain[12][6];  /// for each direction, lists the 6 possible neighbors
extern int pattern_h[24][4][3];   /// precompute the periodic pattern for helix
extern int pattern_h_dir[24][3];   /// for each pattern, specfies the patterns direction
extern int pattern_s[24][2][3];   /// precompute the periodic pattern for strand
extern int pattern_s_dir[24][3];   /// for each pattern, specfies the patterns direction
extern int* code_helix; // for each aa, sets a code: 0= no sec structure, n= number of belonging helix
extern int* code_strand; // for each aa, sets a code: 0= no sec structure, n= number of belonging strand
extern int nhelix;
extern int nstrand;
extern int helix[100][2];
extern int strand[100][2];

extern int discrepancy;

int lastExpandedNodes=0;
int minLev=n;

int temppos[1000][3];  // for bound estimation


void search_best_fit(int lev)
{
     /// implements simple search with backtrack
     /// assume that constraint propagation is applied on level 0 (initial state)

  // keep choice point info in the C stack
  int trailtop;  // top of trail at the start
  int variable;  // variable picked for labeling
  int llimit[3]; // original lower limit for labeled var
  int hlimit[3]; // original higher limit for the labeled var
  int choice[3]; // next value to be given
     
  int go_on;
  int flag;
  int done;
  
  int prev1;
  int prev2;
  int fast_dir; /// if -1 no fast domain possible (use normal domain enumeration), if >=0 apply fast in_domain
  int fast_dir_ct=0;
  int fast_choice1,fast_choice2,fast_choice3;


  /// base cases     
  if (lev<1) {
	  printf("ERROR: labeling called with %d\n",lev);
	  fflush(stdout);
	  return;
  }
  
  if (minLev>lev) minLev=lev;

  if (SEARCH_DBG) {
	  printf("Level %d\n",lev);
	  fflush(stdout);
  }
  
        const double rescaleFactor = 3.8/sqrt(2);
  
  if (lev>n)
  {

        // Energy (by hand --- inefficient)
        double energy=0;
//        const double rescaleFactor = 1.0;
        int ct=0;
        for (int i=0;i<n;i++){
         for (int j=i+1;j<n;j++)
		 {
            double dx,dy,dz;
            dx=STATE->variables[i].l[0]-STATE->variables[j].l[0];
            dy=STATE->variables[i].l[1]-STATE->variables[j].l[1];
            dz=STATE->variables[i].l[2]-STATE->variables[j].l[2];
            double d=sqrt(dx*dx+dy*dy+dz*dz)*rescaleFactor;
            energy+=(dist[i][j]-d)*(dist[i][j]-d);
            ct++;
         }         
        }

        
        energy=sqrt(energy/ct);

		if (best_energy>energy) 
           {
            for (int i=0;i<n;i++)
            //printf("%d %d %d\n",STATE->variables[i].l[0],STATE->variables[i].l[2],STATE->variables[i].l[2]);fflush(stdout);
                                 
             best_energy=energy;
             allvar_fill_from(best_sol,STATE);
             printf("Best new minimum: %f. Leaves %ld, Nodes %ld\n",best_energy,leaves,expanded_nodes);
             fflush(stdout);
             search_print_result_best(best_sol,0);

		 }
		 //else{
         //    printf("NON Best new minimum: %f. Leaves %ld, Nodes %ld\n",best_energy,leaves,expanded_nodes);
		 //}

		// count number of original positions used
/*		ct=0;
        for (int i=0;i<n;i++){
		  if (STATE->variables[i].l[0]==refFCC[i][0] && 
		      STATE->variables[i].l[1]==refFCC[i][1] && 
		      STATE->variables[i].l[2]==refFCC[i][2])
			  ct++;
		}
		printf("Expanded %d nodes: diff %d\n",expanded_nodes,n-ct);
		fflush(stdout);
*/		

		leaves++;
       return;
  }

  
 
  
/////////////////////// bound estimation: inefficient, can save some work!
/*
        int ct=0;
		double estimate=0;

		
        for (int i=0;i<lev;i++){  // already labelled
		temppos[i][0]=STATE->variables[i].l[0];
        temppos[i][1]=STATE->variables[i].l[1];
        temppos[i][2]=STATE->variables[i].l[2];
         for (int j=0;j<i;j++)
		 {
            double dx,dy,dz;
            dx=STATE->variables[i].l[0]-STATE->variables[j].l[0];
            dy=STATE->variables[i].l[1]-STATE->variables[j].l[1];
            dz=STATE->variables[i].l[2]-STATE->variables[j].l[2];
            double d=sqrt(dx*dx+dy*dy+dz*dz)*rescaleFactor;
            estimate+=(dist[i][j]-d)*(dist[i][j]-d);
            ct++;
         }         
        }
		
		
        for (int i=lev;i<n;i++){
		    // for each non labelled i
			// test greedy best position with j < i contribution
			
            double best=-1;
			int bestct=0;
			int bestidx=-1;
			for (int px=STATE->variables[i].l[0];px<=STATE->variables[i].h[0];px++)
			for (int py=STATE->variables[i].l[1];py<=STATE->variables[i].h[1];py++)
			for (int pz=STATE->variables[i].l[2];pz<=STATE->variables[i].h[2];pz++)
			if ((px+py+pz)%2==0){  // on fcc lattice{
			
			  double contr=0;  // best contribution pxpypz - with any i-3 + i-2 points
			  int lct=0;
              for (int j=0;j<i;j++){		 			   
				  int dx=px-temppos[j][0];
				  int dy=py-temppos[j][1];
				  int dz=pz-temppos[j][2];
                  double d=sqrt(dx*dx+dy*dy+dz*dz)*rescaleFactor;
                  double contr1=(dist[i][j]-d)*(dist[i][j]-d);				
			      contr+=contr1;
				  lct++;
			  }
			  if (best<0 || best>contr){
			   best=contr;
			   bestct=lct;
			   temppos[i][0]=px;    // load new greedy position
			   temppos[i][1]=py;
			   temppos[i][2]=pz;
			  }
			}  
			
			estimate+=best;
			ct+=bestct;			
		}

        estimate=sqrt(estimate/ct);
		
		//if (lev<20) printf("%d: %f estimate\n",lev,estimate);fflush(stdout);

		if (estimate>best_energy) // verify bound
		  return;
*/		
		
//	     if (expanded_nodes-lastExpandedNodes>10000){
//		   printf("highest depth in last 10000 nodes expanded: %d\n",minLev);fflush(stdout);
//		   minLev=n;
//		   lastExpandedNodes=expanded_nodes;
//		  }









  
  /// inductive step
       // set up the current level 
       trailtop = trailstack->count;

       /// var_selection
       variable = var_on_lev[lev] = search_variable_selection(lev,STATE); //look in the current status (in last explored level)
       
       memcpy(llimit,STATE->variables[variable].l,sizeof(int)*3);
       memcpy(hlimit,STATE->variables[variable].h,sizeof(int)*3);
       memcpy(choice,STATE->variables[variable].l,sizeof(int)*3);

//////////////////////////////////////////
//// check if it's possible to start a fast indomain
       fast_dir=-1;
       prev1=-1;
       // when two vars on left are labelled
       if (variable-1>=0 &&
           STATE->variables[variable-1].labelled){
             prev1=variable-1;
       fast_dir=0;
       fast_dir_ct=-1;

               done=0;
               while (!done)
	           {
      	          if (fast_dir_ct>=12)
      	            done = 1;
			      else{
			        if (fast_dir_ct==-1){  // special case: original value first (also checks that that there is the way to connect)
				     fast_choice1=refFCC[variable][0];
				     fast_choice2=refFCC[variable][1];
				     fast_choice3=refFCC[variable][2];
					 // is it possible to reach the point from last labelled: optimized version, just coordinate test!
					 if (abs(fast_choice1-STATE->variables[prev1].l[0])+
					     abs(fast_choice2-STATE->variables[prev1].l[1])+
					     abs(fast_choice3-STATE->variables[prev1].l[2])==2 &&
						 llimit[0]<=fast_choice1 &&  // and in the domain
                         hlimit[0]>=fast_choice1 &&
                         llimit[1]<=fast_choice2 &&
                         hlimit[1]>=fast_choice2 &&
                         llimit[2]<=fast_choice3 &&
                         hlimit[2]>=fast_choice3)
					     done=1;					 
				    }
    	            else
                    {  
                      fast_choice1=STATE->variables[prev1].l[0]+directions[fast_dir_ct][0];
                      fast_choice2=STATE->variables[prev1].l[1]+directions[fast_dir_ct][1];
                      fast_choice3=STATE->variables[prev1].l[2]+directions[fast_dir_ct][2];
                      if (SEARCH_DBG) {
                    	  printf("test %d %d %d: %d %d %d - %d %d %d\n",
							 fast_choice1,
							 fast_choice2,
							 fast_choice3,
							 llimit[0],llimit[1],llimit[2],
							 hlimit[0],hlimit[1],hlimit[2]);
                    	  fflush(stdout);
                      }
                      /// intersect with domain on var...
                      if (llimit[0]<=fast_choice1 &&
                          hlimit[0]>=fast_choice1 &&
                          llimit[1]<=fast_choice2 &&
                          hlimit[1]>=fast_choice2 &&
                          llimit[2]<=fast_choice3 &&
                          hlimit[2]>=fast_choice3 &&
						  (fast_choice1 != refFCC[variable][0] ||   // skip the original position here
						   fast_choice2 != refFCC[variable][1] ||
						   fast_choice3 != refFCC[variable][2]))
      	                  done = 1;  /// fast_dir_ct generates an ok point      	          
                    }				  
				    fast_dir_ct++;
				  }
                  
               } 
       }           
//////////////////////////////////////////
//// prepare data for normal indomain
       else{       
        // ensure correct value
        done=0;
        while (!done)
	    {
	    if ( (choice[0]+choice[1]+choice[2]) % 2 == 0)
	      done = 1;
	    else
	      {
	      choice[0]++;
	      if (choice[0] > hlimit[0])
	        {choice[0] = llimit[0]; choice[1]++;}
	      if (choice[1] > hlimit[1])
	        {choice[1] = llimit[1]; choice[2]++;}
	      if (choice[2] > hlimit[2])
	        {choice[0] = -1;done = 1;}
	      }
	    }
       } 
       //allvar_print_stat(STATE);
       
//////////////////////////////////////////
//// assign every point in the domain and open branches
       go_on=1;
       while(go_on)  // copy new level and sets variable value
       {
         if (fast_dir<0)    
      /// normal indomain	     
	     {
	      if (choice[0] == -1)
	        go_on = 0;
	      else
	      {
	        var_init_point(STATE->variables+variable,choice[0],choice[1],choice[2]);
	        if (SEARCH_DBG) {
	        	printf("\t Lev%d: Trying value %d %d %d for var%d\n",lev,choice[0],choice[1],choice[2],variable);
	        	fflush(stdout);
	        }
  	        //if (SEARCH_DBG) allvar_print_stat(STATE);
	        STATE->variables[variable].labelled = 1;
 	        done = 0;
 	        //move to next value in the choice point
 	        while (!done)
	        {
		      choice[0]++;
		      if (choice[0] > hlimit[0])
		        {choice[0] = llimit[0];choice[1]++;}
  		      if (choice[1] > hlimit[1])
	            {choice[1] = llimit[1];choice[2]++;}
		      if (choice[2] > hlimit[2])
		        {choice[0] = -1;done = 1;}
		      else
              if ( (choice[0]+choice[1]+choice[2])%2 == 0)
		        done = 1;
	        }
          } 
	     }
	     else 
      /// use fast indomain
	     {
             if (fast_dir_ct>=12)   /// no more directions to explore
	          go_on = 0;                                
             else
             {
              //assign
	           var_init_point(STATE->variables+variable,fast_choice1,fast_choice2,fast_choice3);
	           if (SEARCH_DBG) {
	        	   printf("\t Lev%d: FAST%d: Trying value %d %d %d for var%d\n",lev,fast_dir_ct,fast_choice1,fast_choice2,fast_choice3,variable);
	        	   fflush(stdout);
	           }
  	           //if (SEARCH_DBG) allvar_print_stat(STATE);
	           STATE->variables[variable].labelled = 1;
			   
			   
               done=0;
               while (!done)
	           {
      	          if (fast_dir_ct>=12)
      	            done = 1;
			      else{
			        if (fast_dir_ct==-1){  // special case: original value first (also checks that that there is the way to connect)
				     fast_choice1=refFCC[variable][0];
				     fast_choice2=refFCC[variable][1];
				     fast_choice3=refFCC[variable][2];
					 // is it possible to reach the point from last labelled: optimized version, just coordinate test!
					 if (abs(fast_choice1-STATE->variables[prev1].l[0])+
					     abs(fast_choice2-STATE->variables[prev1].l[1])+
					     abs(fast_choice3-STATE->variables[prev1].l[2])==2 &&
						 llimit[0]<=fast_choice1 &&  // and in the domain
                         hlimit[0]>=fast_choice1 &&
                         llimit[1]<=fast_choice2 &&
                         hlimit[1]>=fast_choice2 &&
                         llimit[2]<=fast_choice3 &&
                         hlimit[2]>=fast_choice3)
					     done=1;					 
				    }
    	            else
                    {  
                      fast_choice1=STATE->variables[prev1].l[0]+directions[fast_dir_ct][0];
                      fast_choice2=STATE->variables[prev1].l[1]+directions[fast_dir_ct][1];
                      fast_choice3=STATE->variables[prev1].l[2]+directions[fast_dir_ct][2];
                      if (SEARCH_DBG) {
                    	  printf("test %d %d %d: %d %d %d - %d %d %d\n",
							 fast_choice1,
							 fast_choice2,
							 fast_choice3,
							 llimit[0],llimit[1],llimit[2],
							 hlimit[0],hlimit[1],hlimit[2]);
                    	  fflush(stdout);
                      }
                      /// intersect with domain on var...
                      if (llimit[0]<=fast_choice1 &&
                          hlimit[0]>=fast_choice1 &&
                          llimit[1]<=fast_choice2 &&
                          hlimit[1]>=fast_choice2 &&
                          llimit[2]<=fast_choice3 &&
                          hlimit[2]>=fast_choice3 &&
						  (fast_choice1 != refFCC[variable][0] ||   // skip the original position here
						   fast_choice2 != refFCC[variable][1] ||
						   fast_choice3 != refFCC[variable][2]))
      	                  done = 1;  /// fast_dir_ct generates an ok point      	          
                    }				  
				    fast_dir_ct++;
				  }
                  
               } 
             }
         } /// the new domain has been assigned

         // count number of assignments different from original
         int ct=0;
         for (int i=0;i<lev;i++){
         if (STATE->variables[i].l[0]!=refFCC[i][0] ||
	         STATE->variables[i].l[1]!=refFCC[i][1] ||
	         STATE->variables[i].l[2]!=refFCC[i][2]){
		     //if (SEARCH_DBG) 
		     /*printf("%d: %d %d %d - %d %d %d\n",
		       i,
		       STATE->variables[i].l[0],
		       STATE->variables[i].l[1],
		       STATE->variables[i].l[2],
		       refFCC[i][0],
		       refFCC[i][1],
		       refFCC[i][2]);
			  */
		       ct++;
     		}
         }
  
         if (ct>discrepancy) {
         //if (SEARCH_DBG) 
	     //printf("Discrepancy: %d\n",ct);
         go_on=0;
         }
	     
	     if (go_on)
	     {           
  
	       expanded_nodes++;
	       flag=cstore_consistent(STATE,var_on_lev[lev]);
	       if (flag) // check consistency
	       {
	         flag=cstore_propagate(STATE,var_on_lev[lev],trailtop);
		     if (flag)
		     {
                  search_best_fit(lev+1);		 
             } 
	       }
	       
	     }
	     backtrack(lev,trailtop,STATE);
       } /// loop while
       
       STATE->variables[variable].labelled = 0;
       memcpy(STATE->variables[variable].l,llimit,sizeof(int)*3);
       memcpy(STATE->variables[variable].h,hlimit,sizeof(int)*3);
       if (SEARCH_DBG) { printf("Backtrack lev %d\n",lev);fflush(stdout);}
       if (SEARCH_DBG) {
    	   printf("\t RESTORE bounds for var %d to %d %d %d  - %d %d %d\n",var_on_lev[lev],
       	      STATE->variables[variable].l[0],
              STATE->variables[variable].l[1],
              STATE->variables[variable].l[2],
              STATE->variables[variable].h[0],
              STATE->variables[variable].h[1],
              STATE->variables[variable].h[2]);
    	   fflush(stdout);
       }
        //if (SEARCH_DBG) allvar_print_stat(STATE);
    
     if (lev==1) 
      {printf("End of search\n"); fflush(stdout);
      }
}

