/*
      
*/

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>
#include "AllVariables.h"
#include "ConstraintStore.h"
#include "Search.h"
#include "User.h"

using namespace std;

char prot_name[100];
int n; // number of variables
char primary[1000]; // sequence
char prim[2][1000];   /// data for XYZ proteins
double ter[2][1000][3];
int refFCC[1000][3];
double** dist;
int len[2];
int leny; ///// length of coil y in xyz

int* code_strand; // for each aa, sets a code: 0= no strand, n= number of belonging strand
int* code_helix; // for each aa, sets a code: 0= no strand, n= number of belonging helix

int nhelix=0;
int nstrand=0;
int nssbond=0;
int helix[100][2];
int strand[100][2];
int ssbond[100][2];

int minSSE;    /// stores the first aa involved in some SSE (used to compute a better FF strategy)
int maxSSE;    /// stores the last aa involved in some SSE 

int discrepancy=0;

///// conversion utility
char cv31(char* a)
{
    char a1='!';
    if (strcmp("ALA",a)==0) a1= 'a';
    if (strcmp("ARG",a)==0) a1= 'r';
    if (strcmp("ASN",a)==0) a1= 'n';
    if (strcmp("ASP",a)==0) a1= 'd';
    if (strcmp("CYS",a)==0) a1= 'c';
    if (strcmp("GLN",a)==0) a1= 'q';
    if (strcmp("GLU",a)==0) a1= 'e';
    if (strcmp("GLY",a)==0) a1= 'g';
    if (strcmp("HIS",a)==0) a1= 'h';
    if (strcmp("ILE",a)==0) a1= 'i';
    if (strcmp("LEU",a)==0) a1= 'l';
    if (strcmp("LYS",a)==0) a1= 'k';
    if (strcmp("MET",a)==0) a1= 'm';
    if (strcmp("PHE",a)==0) a1= 'f';
    if (strcmp("PRO",a)==0) a1= 'p';
    if (strcmp("SER",a)==0) a1= 's';
    if (strcmp("THR",a)==0) a1= 't';
    if (strcmp("TRP",a)==0) a1= 'w';
    if (strcmp("TYR",a)==0) a1= 'y';
    if (strcmp("VAL",a)==0) a1= 'v';
    return a1;
} 






/// add constraints for a generic protein
void protein_fit(AllVariables* vars, int d_size)
{
     
    /// next constraint for consecutive amino acids 
    for (int i=0;i<n-1;i++)
      cstore_add_next(i,i+1);
      
    /// non occupancy constraints
    for (int i=0;i<n;i++)
    for (int j=i+2;j<n;j++)
    {
         ////  no volume interaction (stronger version of all_different)
        cstore_add(CONSTR_EUCL_G,i,j,1);  ///  every pair of non next points (on primary), are at distance
        cstore_add(CONSTR_EUCL_G,j,i,1);  /// greater than a lattice unit ( sqrt(4) is the first allowed eucl distance on lattice)
    } 
        
    // setup first point
        
    int point[3];
    int point2[3];
    point[0]=128;
    point[1]=128;
    point[2]=128;    
    int var=0;
    var_init_point(vars->variables+var,point);
    
    // for each remaining variable assign a domain centered on the
    // original poisition (greedy solution) + box
    
    int d=d_size;
    
    for (int i=1;i<n;i++){
     for (int z=0;z<3;z++){
      point[z]=-d+refFCC[i][z];point2[z]=d+refFCC[i][z];
      printf("%d %d, ",point[z],point2[z]);
      }
     printf("\n");
     var_init_bounds(vars->variables+var+i,point,point2);
    } 

///// secondary structure handling (none!)
    code_helix=(int*)calloc(n,sizeof(int));
    code_strand=(int*)calloc(n,sizeof(int));
    
    minSSE=n-1;
    maxSSE=0;
    
    for (int i=0;i<n;i++)
    {
      code_helix[i]=0;
      code_strand[i]=0;
    }



}


    



int load_fcc(char* id)
{
     char a[1000];
     char c;
     int v1,v2;
     char name[10];
     int found=0;
     strcpy(prot_name,id);
     
     ifstream in;
     in.open (id, ifstream::in);
             
     int ct=0;
     int x,y,z;          
     while (!in.eof())
     {
       in >> x >> y >> z;
       if (!in.eof()){
       refFCC[ct][0]=x+128;
       refFCC[ct][1]=y+128;
       refFCC[ct][2]=z+128;
       a[ct]='a';
       ct++;
       }
     }
     in.close();
     a[ct]=0;
     strcpy(primary,a);
     n=ct;
     return found;
}



//OK!
void read_atom_pdb(char *buf, char *at_rec, int *p_n_at_pdb,
   char *at_name, char *alt_loc, char *res_name, char *chain,
   int *p_n_res, char *res_ins, float *p_x, float *p_y, float *p_z,
   float *p_occ, float *p_temp_f, int *p_foo_num) {


        char tok[10];


        strncpy(tok,buf,5);
        tok[5] = '\0';
        sscanf(tok,"%s", at_rec);


        strncpy(tok,buf + 6,6);
        tok[6] = '\0';
        sscanf(tok,"%i",p_n_at_pdb);


        strncpy(tok,buf + 12,5);
        tok[5] = '\0';
        sscanf(tok,"%s", at_name);


        strncpy(tok,buf + 16,1);
        tok[1] = '\0';
        if(sscanf(tok,"%s", alt_loc) == -1) strcpy(alt_loc," ");


        strncpy(tok,buf + 17,3);
        tok[3] = '\0';
        sscanf(tok,"%s", res_name);


        strncpy(tok,buf + 21,1);
        tok[1] = '\0';
        if(sscanf(tok,"%s", chain) == EOF) strcpy(chain," ");


        strncpy(tok,buf + 22,5);
        tok[5] = '\0';
        sscanf(tok,"%i", p_n_res);


        strncpy(tok,buf + 26,1);
        tok[1] = '\0';
        if (sscanf(tok,"%s", res_ins) == EOF) strcpy(res_ins," ");


        strncpy(tok,buf + 30,8);
        tok[8] = '\0';
        sscanf(tok,"%f", p_x);
        
        strncpy(tok,buf + 38,8);
        tok[8] = '\0';
        sscanf(tok,"%f", p_y);
        
        strncpy(tok,buf + 46,8);
        tok[8] = '\0';
        sscanf(tok,"%f", p_z);


        strncpy(tok,buf + 54,6);
        tok[6] = '\0';
        sscanf(tok,"%f", p_occ);


        strncpy(tok,buf + 60,6);
        tok[6] = '\0';
        sscanf(tok,"%f", p_temp_f);


        strncpy(tok,buf + 67,3);
        tok[3] = '\0';
        sscanf(tok,"%i", p_foo_num);


}


int load_pdb(char* toopen, int which){
    FILE* fi;

    int ofs_res=-1;
    int ct=0;
	 
      char buf[121];
      char at_rec[5], at_name[5], alt_loc[2], res_name[4],
           chain[2], res_ins[2];
      int n_at_pdb, n_res, foo_num;
      float x, y, z, temp_f, occ;
        

    fi=fopen(toopen,"r");

    if (fi==NULL) {
	    std::cerr << "[File error, skip " << toopen << " %s]\n";
	    return -1;
	 }

    int stop=0;

    const double rescaleFactor = 1.0;
//    const double rescaleFactor = 3.8/sqrt(2);
       
    while(fgets(buf,120,fi) != NULL &&  stop<2){
	   //printf("%s\n",buf); 
      if((!strncmp("ATOM",buf,4) ||
          !strncmp("HETATM",buf,6)
          )
         && stop<2) 
      {  
        stop=1;
        read_atom_pdb(buf, at_rec, &n_at_pdb, at_name, alt_loc,
                      res_name, chain, &n_res, res_ins, &x, &y, &z, 
                      &occ, &temp_f, &foo_num);

       char at_single_name=buf[13];
       // no H
       if (chain[0]=='P' && (
           //at_single_name=='H' ||
           at_single_name=='C' ||
           at_single_name=='S' ||
           at_single_name=='N' ||
           at_single_name=='O' ||
           at_single_name=='P'))
       {
	
      if (1==1){
       printf("%d ",n_res);
       printf("%s %c ",at_name,cv31(res_name));
       printf("%s %s",res_name,chain);
//       printf("%c,",cv(res_name));
	    printf("%f ",x);
	    printf("%f ",y);
	    printf("%f\n",z);
	    }

      if (ofs_res==-1) {
	      ofs_res=n_res;      
	      }
      
      int i=n_res-ofs_res;
      primary[i]=cv31(res_name);
      ter[which][i][0]=(x/rescaleFactor);
      ter[which][i][1]=(y/rescaleFactor);
      ter[which][i][2]=(z/rescaleFactor);
//      ter[which][i][0]=(int)(x/3.8*sqrt(2));
//      ter[which][i][1]=(int)(y/3.8*sqrt(2));
//      ter[which][i][2]=(int)(z/3.8*sqrt(2));
      ct=i;
      //printf("Read: %d %c %d %d %d\n",i,prim[which][i],ter[which][i][0],ter[which][i][1],ter[which][i][2]);
      }	 
     }//if strncmp
     else
      {
	      if (stop==1) 
	      {stop=2;
          //printf("Protein done: start %d\n",ofs_res);
	      }
	   }

    }

     len[which]=ct+1;
     printf("Prot len:%d\n",len[which]);     

//	 printf(" closing!\n");
	 if (fi!=NULL) 
	 {
		fclose(fi);		
		//printf("load ok\n");
    }


   // precompute distances
   dist=(double**)calloc(n,sizeof(double*));
   for (int i=0;i<n;i++){
       dist[i]=(double*)calloc(n,sizeof(double));
       for (int j=0;j<n;j++){       
           double x,y,z;
           x=ter[0][i][0]-ter[0][j][0];
           y=ter[0][i][1]-ter[0][j][1];
           z=ter[0][i][2]-ter[0][j][2];
           dist[i][j]=sqrt(x*x+y*y+z*z);
       }
   }
    
    
   return ofs_res;

}





int main(int argc, char *argv[])
{         
     
 if (argc!=5)
      {
      printf("Usage: %s FCC-positions original-protein.pdb box-size discrepancy\n",argv[0]);
      #if (defined(_WIN32) || defined(__WIN32__) || defined(WIN32))
      system("PAUSE");
      #endif
      return 1;
      }
    printf("Constraint solver for Lattice\n");
        
    // load protein
    load_fcc(argv[1]);
    load_pdb(argv[2],0);
          
    AllVariables* vars=(AllVariables*)malloc(sizeof(AllVariables));
    
    allvar_init(vars);
    printf("Primary: %s\n",primary);
    printf("Length: %d\n",n);       

    //init constraint store  
    cstore_init(STORE_AC3);
   
    protein_fit(vars,atoi(argv[3]));
    //cstore_list();

	discrepancy=atoi(argv[4]);
	
  /// launch search
      search(vars,SEARCH_BEST_FIT, SEARCH_FF);

    cstore_stat(); 
#if (defined(_WIN32) || defined(__WIN32__) || defined(WIN32))
    system("PAUSE");
#endif
    return 0;
}
